import 'dotenv/config';

import { createLog } from '../log';
import type { Logger } from 'winston';
import { Database } from 'sqlite3';
import { envString } from '../env';

interface LatLng {
	latitude : number;
	longitude : number;
}

interface Photo {
	[key: string] : any;
	id : string;
	name : string;
	description : string;
	original? : string;
	hash : string;
	width : number;
	height : number;
	thumbHeight? : number;
	format? : string;
	filmstock? : string;
	location? : string;
	latitude? : number;
	longitude? : number;
	discovered ? : number;
	created? : number;
	updated? : number;
	bsky? : boolean;
	deleted? : boolean;
	score? : number;
}

export class DB {
	private db : Database;
	private log : Logger;

	constructor () {
		const databaseFile : string = envString('DB', 'data/site.db');
		this.log = createLog('db');
		this.db = new Database(databaseFile);
		this.db.run( 'PRAGMA journal_mode = WAL;');
		this.log.info(`DB: ${databaseFile}`);
	}

	private async run (query : string, args : any[] = []) {
		return new Promise((resolve : Function, reject : Function) => {
			return this.db.run(query, args, (err : Error, rows : any[]) => {
				if (err) return reject(err);
				return resolve(true);
			});
		});
	}

	private async all (query : string, args : any[] = []) : Promise<any[]> {
		return new Promise((resolve : Function, reject : Function) => {
			return this.db.all(query, args, (err : Error, rows : any[]) => {
				if (err) return reject(err);
				return resolve(rows);
			});
		});
	}

	private toBoolean (val : number) : boolean {
		return val === 1 ? true : false;
	}

	private fromBoolean (val : boolean) : number {
		return val ? 1 : 0;
	}

	//CASE WHEN LOWER(active) = 'true' THEN 1 ELSE 0 END AS active_bool
	public async create (photo : Photo) {
		const keys : string[] = Object.keys(photo);
		const query : string = `INSERT INTO photos (${keys.join(',')}) VALUES (${keys.map(el=>'?').join(',')});`;
		const values : any[] = [];
		for (let key of keys) {
			if (typeof (photo as any)[key] === 'boolean') {
				values.push( this.fromBoolean( (photo as any)[key] ));
			} else {
				values.push((photo as any)[key]);
			}
		}
		try {
			await this.run(query, values);
			this.log.info(`Inserted new photo ${photo.name}`);
		} catch (err) {
			throw err;
		}
	}

	public async existsName (name : string) : Promise<boolean> {
		const query : string = `SELECT id FROM photos WHERE name = ? LIMIT 1;`;
		let rows : any[] = [];
		let exists : boolean = false;
		try {
			rows = await this.all(query, [ name ]);
		} catch (err) {
			this.log.error(`Error finding photo by name ${name}`, err);
		}
		if (rows.length > 0) {
			exists = true;
		}
		return exists;
	}

	public async existsHash (hash : string) : Promise<boolean> {
		const query : string = `SELECT id FROM photos WHERE hash = ? LIMIT 1;`;
		let rows : any[] = [];
		let exists : boolean = false;
		try {
			rows = await this.all(query, [ hash ]);
		} catch (err) {
			this.log.error(`Error finding photo by hash ${hash}`, err);
		}
		if (rows.length > 0) {
			exists = true;
		}
		return exists;
	}

	public async getAll () : Promise<Photo[]> {
		const query : string = `SELECT * FROM photos ORDER BY score ASC, created DESC;`;
		let rows : Photo[] = [];
		try {
			rows = await this.all(query);
		} catch (err) {
			this.log.error(`Error getting all photos`, err);
		}
		return rows;
	}

	public async update (photo : Photo, fields : string[]) {
		if (fields.indexOf('updated') === -1) {
			fields.push('updated');
			photo.updated = Date.now();
		}
		const values : any[] = fields.map(el => photo[el]);
		const query : string = `UPDATE photos SET ${fields.map(el => { return el + ' = ?'}).join(',')} WHERE id = ?;`;
		values.push(photo.id);
		try {
			await this.run(query, values);
		} catch (err) {
			this.log.error(`Error updating photo with columns: ${fields.join(', ')}`, err);
			this.log.info(query);
		}
	}

	public async getBsky (random : boolean = false) : Promise<Photo> {
		const order : string = random ? 'RANDOM()' : 'score ASC, created DESC';
		const query : string = `SELECT * FROM photos WHERE bsky = 0 ORDER BY ${order} LIMIT 1;`;
		let rows : Photo[] = [];
		try {
			rows = await this.all(query);
		} catch (err) {
			this.log.error(`Error getting all photos`, err);
		}
		if (rows.length === 0) {
			return null;
		}
		return rows[0];
	}

	public async confirmBsky (id : string) {
		const query : string = `UPDATE photos SET bsky = 1 WHERE id = ?;`;
		
		try {
			await this.run(query, [ id ]);
		} catch (err) {
			throw err;
		}
	}

	public async cacheLocation (location : string, latlng : LatLng) {
		const query : string = `INSERT OR IGNORE INTO geocode (location, latitude, longitude) VALUES (?, ?, ?);`;
		try {
			await this.run(query, [location, latlng.latitude, latlng.longitude]);
		} catch (err) {
			//ignore
		}
	}

	public async getLocation (location : string) : Promise<LatLng> {
		const query : string  = `SELECT latitude, longitude FROM geocode WHERE location = ? LIMIT 1;`;
		let rows : any[] = [];
		let res : LatLng = null;
		try {
			rows = await this.all(query, [location]);
		} catch (err) {
			//ignore
		}

		if (rows.length > 0) {
			res = {
				latitude : rows[0].latitude,
				longitude : rows[0].longitude
			}
		}

		return res;
	}

}

module.exports = { DB };
export type { Photo, LatLng };