import 'dotenv/config';

import { createLog } from './log';
import type { Logger } from 'winston';
import { readFile, unlink } from 'fs/promises';
import { join } from 'path';
import { BskyAgent } from '@atproto/api';
import type CallOptions from '@atproto/api';
import { ArgumentParser } from 'argparse';
import moment from 'moment';
import type { Moment } from 'moment';

import { DB } from './db';
import type { Photo } from './db';
import { envString } from './env';
import { size } from './size';
import { tmp } from './tmp';
import { Shell } from './shell';

class Post {
	private log : Logger = createLog('post');
	private db : DB = new DB();
	private bskyAgent : BskyAgent = new BskyAgent({ service: 'https://bsky.social' });
	private random : boolean = false;

	constructor () {
		const parser = new ArgumentParser({
			description: 'Generate script'
		});
		parser.add_argument('-r', '--random', { action : 'store_true', default : false, help: 'Choose one at random' });
		const args : any = parser.parse_args();
		this.random = args.random;
		this.post();
	}

	private async post () {
		await this.bsky();
	}

	private bskyDescription (photo : Photo) : string {
		let str : string = `${photo.description}

${photo.format}${photo.filmstock !== 'Unknown' ? ' - ' + photo.filmstock : ''}`;
		const created : Moment = moment.unix(photo.created / 1000);
		const now : Moment = moment();
		if (now.format('YYYY') !== created.format('YYYY')) {
			str += '\n' + created.format('MMMM YYYY');
		}
		return str;
	}

	private async bsky () {
		const photo : Photo = await this.db.getBsky(this.random);
		const identifier : string = envString('BSKY_USER', null);
		const password : string = envString('BSKY_PASSWORD', null);
		const maxSize : number = 976560;
		let text : string;
		let post : any;
		let imagePath : string;
		let imageData : Buffer;
		let imageSize : number;
		let blobResponse : any;
		let postData : any;
		let usedTmp : boolean = false;

		if (identifier === null || password === null) {
			this.log.error(`Cannot post without Bluesky username or password`);
			return;
		}

		if (photo === null) {
			this.log.warn(`Nothing to post`);
			return;
		}

		try {
			await this.bskyAgent.login({ identifier, password });
			this.log.info(`Logged into Bluesky`);
		} catch (err) {
			this.log.error(`Error logging into Bluesky`, err);
			return;
		}

		text = this.bskyDescription(photo);
/*
                        embed: {
                            images: [
                                {
                                    image: testUpload.data.blob,
                                    alt: "",
                                },
                            ],
                            $type: "app.bsky.embed.images",
                        },
*/
		this.log.info(text);
		
		imagePath = join(envString('WWW', './www'), 'img', `${photo.id}_full.jpg`);

		try {
			imageSize = await size(imagePath);
		} catch (err) {
			this.log.error(`Error getting image size`, err);
			return;
		}

		//976.56KB limit
		if (imageSize > maxSize) {
			this.log.info(`Image size too large: ${imageSize} > ${maxSize}`);
			
			try {
				imagePath = await this.imgFix(imagePath);
			} catch (err) {
				this.log.error(`Error resizing image with img_fix.sh`, err);
				return;
			}

			usedTmp = true;

			try {
				imageSize = await size(imagePath);
			} catch (err) {
				this.log.error(`Error getting resized image size`, err);
				return;
			}

			this.log.info(`Image resized to: ${imageSize}`);
		}

		try {
			imageData = await readFile(imagePath);
		} catch (err) {
			this.log.error(`Error reading image data ${imagePath}`, err);
			return;
		}

		try {
			blobResponse = await this.bskyAgent.uploadBlob(imageData, { contentType: 'image/jpeg' } as any)
		} catch (err) {
			this.log.error(`Error uploading blob to Bluesky`, err);
			return;
		}

		postData = {
			text,
			embed: {
			$type: 'app.bsky.embed.images',
			images: [{
					alt: photo.description,
					image: blobResponse.data.blob
				}]
			}
		}

		try {
			await this.bskyAgent.post(postData);
			this.log.info(`Posted ${photo.description} (${photo.name})`);
		} catch (err) {
			this.log.error(`Error making Bluesky post`, err);
			return;
		}

		try {
			await this.db.confirmBsky(photo.id);
		} catch (err) {
			this.log.error(`Could not confirm that photo ${photo.name} was posted`, err);
			return;
		}

		if (usedTmp) {
			try {
				await unlink(imagePath);
			} catch (err) {
				this.log.error(`Error removing temp image file ${imagePath}`, err);
				return;
			}
		}

	}

	private async imgFix (imagePath : string) : Promise<string> {
		const dest : string = tmp('.jpg', 'bsky-upload-');
		const args : string[] = ['bash', 'scripts/img_fix.sh', imagePath, dest];
		const shell : Shell = new Shell(args);
	
		await shell.execute();

		return dest;
	}
}

new Post();